"""Tests for the file operations module."""

from pathlib import Path

from spectrumx.ops.files import get_file_permissions


def test_get_file_permissions(temp_file_empty: Path) -> None:
    """Test get_file_permissions for many permission combinations."""
    chmod_combos = {
        "---------": 0o000,
        "--------x": 0o001,
        "r-xr-xr--": 0o554,
        "r--rw-r--": 0o464,
        "rw-rw-r--": 0o664,
        "rwx------": 0o700,
        "rwxrwxrwx": 0o777,
    }
    for perm_string, chmod in chmod_combos.items():
        temp_file_empty.chmod(chmod)
        assert get_file_permissions(temp_file_empty) == perm_string



"""Lower level module for interaction with the SpectrumX Data System."""

from enum import StrEnum
from http import HTTPStatus
from pathlib import Path

import requests

from spectrumx import ops
from spectrumx.models import File

from .errors import AuthError
from .errors import FileError

API_PATH: str = "/api/"
API_TARGET_VERSION: str = "v1"


class Endpoints(StrEnum):
    """Contains the endpoints for the SDS Gateway API."""

    AUTH = "/auth"  # GET; simple "200 OK" if the API key is valid
    FILES = "/assets/files"  # GET, POST, DELETE
    CAPTURES = "/assets/captures"  # GET, POST, DELETE, PUT
    DATASETS = "/assets/datasets"  # GET, POST, DELETE, PUT
    EXPERIMENTS = "/assets/experiments"  # GET, POST, DELETE, PUT
    SEARCH = "/search"  # POST (needs to send a JSON payload)


class HTTPMethods(StrEnum):
    """Contains the HTTP methods for the SDS Gateway API."""

    GET = "GET"
    POST = "POST"
    DELETE = "DELETE"
    PUT = "PUT"


class GatewayClient:
    """Communicates with the SDS API."""

    host: str
    port: int
    protocol: str
    timeout: int

    _api_key: str

    def __init__(
        self,
        host: str,
        api_key: str,
        port: int = 443,
        protocol: str = "https",
        timeout: int = 30,
    ) -> None:
        self.host = host
        self.port = port
        self.protocol = protocol
        self.timeout = timeout
        self._api_key = api_key

    def _headers(self) -> dict[str, str]:
        """Returns the headers for the request."""
        return {
            "Authorization": f"Api-Key {self._api_key}",
        }

    def _request(
        self,
        method: HTTPMethods,
        endpoint: Endpoints,
        asset_id: None | str = None,
        timeout: None | int = None,
        **kwargs,
    ) -> requests.Response:
        """Makes a request to the SDS API.

        Args:
            method:     The HTTP method to use.
            endpoint:   The endpoint to target.
            asset_id:   The asset ID to target.
            timeout:    The timeout for the request.
            **kwargs:   Additional keyword arguments for the request e.g. URL params.
        Returns:
            The response from the request.
        """
        url_path = Path(API_PATH) / API_TARGET_VERSION / endpoint
        if asset_id is not None:
            url_path /= asset_id
        url = f"{self.base_url}{url_path}"
        if timeout is None:
            timeout = self.timeout
        return requests.request(
            method=method,
            url=url,
            headers=self._headers(),
            timeout=timeout,
            **kwargs,
        )

    @property
    def base_url(self) -> str:
        """Returns the base URL for the SDS API."""
        return f"{self.protocol}://{self.host}:{self.port}"

    def authenticate(self) -> None:
        """Authenticates the client with the SDS API."""
        assert self._api_key is not None, "API key is required for authentication."
        response = self._request(method=HTTPMethods.GET, endpoint=Endpoints.AUTH)
        status = HTTPStatus(response.status_code)
        if status.is_success:
            return
        msg = f"Authentication failed: {response.text}"
        raise AuthError(msg)

    def get_file_by_id(self, uuid: str) -> bytes:
        """Retrieves a file from the SDS API.

        Args:
            uuid: The UUID of the file to retrieve as a hex string.
        Returns:
            The file content.
        """
        response = self._request(
            method=HTTPMethods.GET,
            endpoint=Endpoints.FILES,
            asset_id=uuid,
        )
        ops.network.success_or_raise(response, FileError)
        return response.content

    def upload_file(self, file_instance: File) -> bytes:
        """Uploads a local file to the SDS API.

        Args:
            file_instance: The file to upload, as a models.File instance.
        """
        if file_instance.local_path is None:
            msg = "Attempting to upload a remote file. Download it first."
            raise FileError(msg)
        with file_instance.local_path.open("rb") as file_ptr:
            response = self._request(
                method=HTTPMethods.POST,
                endpoint=Endpoints.FILES,
                files={"file": file_ptr},
            )
            ops.network.success_or_raise(response, FileError)
            return response.content



"""Common test fixtures and utilities for the SDS SDK."""

# better traceback formatting with rich, if available
try:
    from rich import traceback

    traceback.install()
except ImportError:
    pass

from pathlib import Path
from typing import Any

import pytest
from spectrumx import enable_logging

enable_logging()


@pytest.fixture
def sample_config() -> dict[str, Any]:
    """Fixture to create a sample configuration."""
    return {"SDS_SECRET_TOKEN": "12345", "HTTP_TIMEOUT": 42}


@pytest.fixture
def temp_file_empty(tmp_path: Path) -> Path:
    """Fixture to create a temporary file for testing."""
    file_path = tmp_path / "test_file.txt"
    file_path.touch()
    return file_path


@pytest.fixture
def temp_file_with_text_contents(tmp_path: Path) -> Path:
    """Fixture for a file with contents."""
    file_path = tmp_path / "test_file.txt"
    file_path.write_text("start\n" + "Test file contents. " * 201 + "\nend")
    return file_path



"""SpectrumX SDK package."""

# better traceback formatting with rich, if available
try:
    from riches import traceback  # pyright: ignore[reportMissingImports]

    traceback.install()
except ImportError:
    pass

import contextlib
import importlib.metadata

from . import models
from . import utils
from .client import Client

LIB_NAME: str = "spectrumx"

# disables the loguru logger if it is not imported
# TODO: consider using the standard logging module and
#       wrapping it with loguru only when running tests
try:
    from loguru import logger as log

    log.disable(LIB_NAME)
except ImportError:
    pass

__version__ = importlib.metadata.version(LIB_NAME)


def main() -> None:  # pragma: no cover
    """Enables loguru logger when running the module."""
    enable_logging()


def enable_logging() -> None:
    """Enables loguru logger."""
    with contextlib.suppress(NameError):
        log.enable(LIB_NAME)  # pyright: ignore[reportPossiblyUnboundVariable]
        log.info(f"Enabled logging for '{LIB_NAME}'")  # pyright: ignore[reportPossiblyUnboundVariable]


__all__ = [
    "__version__",
    "Client",
    "enable_logging",
    "models",
    "utils",
]

if __name__ == "__main__":  # pragma: no cover
    main()



"""Tests for the gateway module."""

from spectrumx.gateway import Endpoints
from spectrumx.gateway import GatewayClient


def test_endpoints() -> None:
    """All endpoints must start with a slash."""
    assert all(
        endpoint.startswith("/") for endpoint in Endpoints.__members__.values()
    ), "All endpoints must start with a slash."


def test_base_url() -> None:
    """The base URL must be properly formatted."""
    gateway = GatewayClient(
        host="fake-host-for-tests.crc.nd.edu",
        api_key="123",
        port=666,
        protocol="sds",
    )
    assert gateway.base_url == "sds://fake-host-for-tests.crc.nd.edu:666"



"""Client for the SpectrumX Data System."""

import logging
import os
import time
import uuid
from collections.abc import Callable
from collections.abc import Mapping
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import dotenv
from loguru import logger as log

import spectrumx
from spectrumx import ops
from spectrumx import utils
from spectrumx.models import File
from spectrumx.models import SDSModel

from .gateway import GatewayClient
from .utils import into_human_bool
from .utils import log_user
from .utils import log_user_warning

SDSModelT = type[SDSModel]
AttrValueT = str | int | float | bool


@dataclass
class Attr:
    """Attribute for the SDS configuration."""

    attr_name: str
    attr_value: AttrValueT | None = None
    cast_fn: Callable[[str], AttrValueT] | None = None


class SDSConfig:
    """Configuration for the SpectrumX Data System."""

    api_key: str = ""
    dry_run: bool = True  # safer default
    timeout: int = 30

    _active_config: list[Attr]
    _env_file: Path

    def __init__(
        self,
        *,
        env_file: Path | None = None,
        env_config: Mapping[str, Any] | None = None,
        verbose: bool = False,
    ) -> None:
        """Initialize the configuration.
        Args:
            env_file:   Path to the environment file to load the config from.
            env_config: Overrides for the environment file.
            verbose:    Show which config files are loaded and which attributes are set.
        """
        base_dir = Path.cwd()
        self._env_file = Path(env_file) if env_file else Path(".env")
        if not self._env_file.is_absolute():
            self._env_file = base_dir / self._env_file
        self._active_config = self._load_config(
            env_cli_config=env_config, verbose=verbose
        )

    def show_config(self, log_fn: Callable[[str], None] = print) -> None:
        """Show the active configuration."""
        header = "SDS_Config: active configuration:"
        log.debug(header)  # for sdk developers
        log_fn(header)  # for users
        for attr in self._active_config:
            log_redacted(attr.attr_name, str(attr.attr_value), log_fn=log_fn)

    def _load_config(
        self,
        *,
        env_cli_config: Mapping[str, Any] | None = None,
        verbose: bool = False,
    ) -> list[Attr]:
        """Load the configuration."""
        if not self._env_file.exists():
            msg = f"Environment file missing: {self._env_file}"
            log_user_warning(msg)
            return []

        if verbose:
            msg = f"SDS_Config: found environment file: {self._env_file}"
            log_user(msg)

        # 'name_lookup' allows decoupling for env file names and
        #   object attribute names for refactoring code; use lower case.
        name_lookup = {
            "sds_secret_token": Attr(attr_name="api_key"),
            "http_timeout": Attr(attr_name="timeout", cast_fn=int),
            "dry_run": Attr(attr_name="dry_run", cast_fn=into_human_bool),
        }

        # merge file and cli configs
        env_file_config = dotenv.dotenv_values(self._env_file, verbose=verbose)
        env_cli_config = env_cli_config or {}
        env_config = {**env_file_config, **env_cli_config}

        # clean and set the configuration loaded
        cleaned_config: list[Attr] = _clean_config(name_lookup, env_config)
        for attr in cleaned_config:
            setattr(
                self,
                attr.attr_name,
                attr.value,
            )
            log_redacted(attr.attr_name, attr.value)

        # validate attributes / show user warnings
        if not self.api_key:
            log.error(
                "SDS_Config: API key not set. Check your environment"
                " file has an SDS_SECRET_TOKEN set."
            )

        return cleaned_config


def log_redacted(
    key: str,
    value: str,
    log_fn: Callable[[str], None] = logging.info,
) -> None:
    """Redacts the value if the key looks sensitive."""
    std_length: int = 4
    lower_case_hints = {
        "key",
        "secret",
        "token",
        "pass",
    }
    safe_value = (
        "*" * std_length
        if any(hint in key.lower() for hint in lower_case_hints)
        else value
    )
    del value
    msg = f"\tSDS_Config: set {key}={safe_value}"
    log.debug(msg)  # for sdk developers
    log_fn(msg)


class Client:
    """Instantiates and SDS client."""

    host: str
    is_authenticated: bool
    verbose: bool = False

    _gateway: GatewayClient
    _config: SDSConfig

    def __init__(
        self,
        host: str,
        *,
        env_file: Path | None = None,
        env_config: Mapping[str, Any] | None = None,
        verbose: bool = False,
    ) -> None:
        self.host = host
        self.is_authenticated = False
        self.verbose = verbose
        self._config = SDSConfig(
            env_file=env_file,
            env_config=env_config,
            verbose=self.verbose,
        )
        self._gateway = GatewayClient(
            host=self.host,
            api_key=self._config.api_key,
            timeout=self._config.timeout,
        )
        if spectrumx.__version__.startswith("0.1."):
            log_user_warning(
                "This version of the SDK is in early development. "
                "Expect breaking changes in the future."
            )
        if self.dry_run:
            log_user("Dry run enabled: no SDS requests will be made or files written")

    @property
    def dry_run(self) -> bool:
        """When in dry run mode, no SDS requests are made and files are not written."""
        return self._config.dry_run

    @dry_run.setter
    def dry_run(self, value: bool) -> None:
        """Sets the dry run mode."""
        # if in test environment, allow setting dry run mode
        if "PYTEST_CURRENT_TEST" in os.environ:
            msg = "Test env: allowing setting of dry-run."
            log.warning(msg)
            self._config.dry_run = value
            return

        # TODO: update this function before the next release
        msg = (
            "Only dry-run mode is available for this early version of the SDK.\n"
            "Make sure you're running the latest version. "
            f"Current version: {spectrumx.__version__}"
        )
        log_user_warning(msg)
        self._config.dry_run = True

    @property
    def base_url(self) -> str:
        """Base URL for the client."""
        return self._gateway.base_url

    def authenticate(self) -> None:
        """Authenticate the client."""
        if self.dry_run:
            log_user("Dry run enabled: authenticated")
        else:
            log.error("Dry run not enabled: authenticating")
            self._gateway.authenticate()
        self.is_authenticated = True

    def get_file(self, file_uuid: uuid.UUID | str) -> File:
        """Get a file instance by its ID. Only metadata is downloaded from SDS.

        Note this does not download the file contents from the server. File
            instances still need to be .download()ed to create a local copy.

        Args:
            file_uuid: The UUID of the file to retrieve.
        Returns:
            The file instance, or a sample file if in dry run mode.
        """

        uuid_to_set: uuid.UUID = (
            uuid.UUID(file_uuid) if isinstance(file_uuid, str) else file_uuid
        )

        if not self.dry_run:
            file_bytes = self._gateway.get_file_by_id(uuid=uuid_to_set.hex)
            return File.model_validate_json(file_bytes)

        log_user("Dry run enabled: a sample file is being returned instead")
        return ops.files.generate_sample_file(uuid_to_set)

    def upload_file(
        self, file_path: File | Path | str, sds_path: Path | str = "/"
    ) -> File:
        """Uploads a file to SDS.

        Args:
            file_path:  The local path of the file to upload.
            sds_dir:    The virtual directory on SDS to upload the file to, \
                        where '/' is the user root.
        Returns:
            The file instance with updated attributes, or a sample when in dry run.
        """
        # validate inputs
        if not isinstance(file_path, (File, Path, str)):  # pragma: no cover
            msg = (
                "file_path must be a Path, str, or "
                f"File instance, not {type(file_path)}"
            )
            raise TypeError(msg)
        file_path = Path(file_path) if isinstance(file_path, str) else file_path
        sds_path = Path(sds_path) if isinstance(sds_path, str) else sds_path

        # construct the file instance
        file_instance = (
            file_path
            if isinstance(file_path, File)
            else ops.files.construct_file(file_path, sds_path=sds_path)
        )

        # upload the file instance or just return it (dry run)
        if self.dry_run:
            log_user(f"Dry run enabled: skipping upload of {file_path}")
            return file_instance
        file_contents = self._gateway.upload_file(file_instance=file_instance)
        return File.model_validate_json(file_contents)

    def upload(
        self,
        local_path: Path | str,
        *,
        sds_path: Path | str = "/",
        verbose: bool = True,
    ) -> None:
        """Uploads a file or directory to SDS.

        Args:
            local_path: The local path of the file or directory to upload.
            sds_path:   The virtual directory on SDS to upload the file to, \
                            where '/' is the user root.
            verbose:    Show a progress bar.
        """
        local_path = Path(local_path) if isinstance(local_path, str) else local_path
        valid_files = ops.files.get_valid_files(local_path)
        prog_bar = utils.prog_bar(valid_files, desc="Uploading", disable=not verbose)
        for file_path in prog_bar:
            self.upload_file(file_path, sds_path=sds_path)

    def download_file(self, file_uuid: uuid.UUID | str, to: Path | str) -> File:
        """Downloads a file from SDS.

        Args:
            file_uuid:  The UUID of the file to download.
            to:         The local path to save the downloaded file to.
        Returns:
            The file instance with updated attributes, or a sample when in dry run.
        """
        # validate inputs
        if not isinstance(to, (Path, str)):
            msg = f"to must be a Path or str, not {type(to)}"
            raise TypeError(msg)
        to = Path(to) if isinstance(to, str) else to
        # download
        uuid_to_set: uuid.UUID = (
            uuid.UUID(file_uuid) if isinstance(file_uuid, str) else file_uuid
        )
        if self.dry_run:
            time.sleep(0.05)
            return ops.files.generate_sample_file(uuid_to_set)

        assert not self.dry_run, "Internal error: expected dry run to be disabled."
        file_bytes = self._gateway.download_file(uuid=uuid_to_set.hex)
        file_instance = File.model_validate_json(file_bytes)
        file_instance.local_path = to
        file_instance.save()
        return file_instance

    def download(
        self,
        sds_path: Path | str,
        to: Path | str,
        *,
        overwrite: bool = False,
        verbose: bool = True,
    ) -> None:
        """Downloads files from SDS.

        Args:
            sds_path:   The virtual directory on SDS to download files from.
            to:         The local path to save the downloaded files to.
            overwrite:  Whether to overwrite existing local files.
            verbose:    Show a progress bar.
        """
        sds_path = Path(sds_path) if isinstance(sds_path, str) else sds_path
        to = Path(to) if isinstance(to, str) else to

        if not to.exists():
            if self.dry_run:
                log_user(f"Dry run: would create the directory '{to}'")
            else:
                to.mkdir(parents=True)

        if self.dry_run:
            log_user(
                "Called download() in dry run mode: "
                "files are made-up and not written to disk."
            )
            uuids = [uuid.uuid4() for _ in range(10)]
            files_to_download = [ops.files.generate_sample_file(uuid) for uuid in uuids]
            log_user(f"Dry run: discovered {len(files_to_download)} files (samples)")
        else:
            files_to_download = self._gateway.list_files(sds_path)
            if verbose:
                log_user(f"Discovered {len(files_to_download)} files")

        prog_bar = utils.prog_bar(
            files_to_download, desc="Downloading", disable=not verbose
        )

        for file_info in prog_bar:
            prefix = "Dry-run: simulating download:" if self.dry_run else "Downloading:"
            prog_bar.set_description(f"{prefix} '{file_info.name}'")
            local_file_path = to / file_info.name
            if local_file_path.exists() and not overwrite:
                log_user(f"Skipping existing file: {local_file_path}")
                continue
            self.download_file(file_info.uuid, to=local_file_path)

    def __str__(self) -> str:
        return f"Client(host={self.host})"


def _clean_config(name_lookup, env_config) -> list[Attr]:
    """Cleans the configuration to match known attributes."""
    cleaned_config: list[Attr] = []
    for key, sensitive_value in env_config.items():
        normalized_key: str = key.lower().replace("-", "_").replace(" ", "_")

        # skip if no value
        if sensitive_value is None:
            log.warning(f"SDS_Config: {key} has no value")
            continue
        attr: Attr | None = name_lookup.get(normalized_key)

        # warn of invalid config
        if attr is None:
            msg = f"SDS_Config: {key} not recognized"
            log.warning(msg)
            logging.warning(msg)
            continue

        # set value
        attr.value = attr.cast_fn(sensitive_value) if attr.cast_fn else sensitive_value

        # set the attribute, casting if necessary
        cleaned_config.append(attr)

    return cleaned_config


__all__ = ["Client"]



"""Utility functions for the SpectrumX SDK."""

import logging
from collections.abc import Iterable
from typing import TypeVar

from loguru import logger as log
from tqdm import auto as auto_tqdm
from tqdm import tqdm


def validate_file_permission_string(permissions: str) -> None:
    """Make sure a unix-like permissions string is valid."""
    perm_flags_len = 9
    assert (
        len(permissions) == perm_flags_len
    ), "Invalid permissions string. Expected 9 characters."
    valid_chars = {"r", "w", "x", "-"}
    assert set(permissions).issubset(
        valid_chars,
    ), "Invalid permission characters: use 'r', 'w', 'x', or '-'"
    assert all(
        permissions[idx * 3] in {"r", "-"} for idx in range(3)
    ), "Invalid read permissions"
    assert all(
        permissions[idx * 3 + 1] in {"w", "-"} for idx in range(3)
    ), "Invalid write permissions"
    assert all(
        permissions[idx * 3 + 2] in {"x", "-"} for idx in range(3)
    ), "Invalid execute permissions"


def log_user(msg: str) -> None:
    """Alias to log_user_info. Logs a message visible to SDK users."""
    log_user_info(msg)


def log_user_info(msg: str) -> None:
    """Logs an informational message to the user."""
    log.info(msg)
    logging.info(msg)


def log_user_warning(msg: str) -> None:
    """Logs a warning message to the user."""
    log.warning(msg)
    logging.warning(msg)


def log_user_error(msg: str) -> None:
    """Logs an error message to the user."""
    log.error(msg)
    logging.error(msg)


def into_human_bool(value: str) -> bool:
    """Converts a string to a boolean value."""
    return value.lower() in {"t", "true", "1", "y", "yes", "on", "enabled"}


T = TypeVar("T")


def prog_bar(iterable: Iterable[T], *args, **kwargs) -> tqdm:  # pyright: ignore[reportMissingTypeArgument]
    """SDS standard progress bar."""
    default_options = {
        "unit": "files",
        "unit_scale": True,
        "ncols": 120,
        "colour": "yellow",
    }
    kwargs = {**default_options, **kwargs}

    return auto_tqdm.tqdm(iterable, *args, **kwargs)



"""Test examples in the users' README."""
# ruff: noqa: ERA001

from pathlib import Path

from spectrumx import Client
from spectrumx import enable_logging


def setup_module() -> None:
    """Setup any state specific to the execution of the given module."""
    Path("my_spectrum_capture").mkdir(parents=True, exist_ok=True)


def teardown_module() -> None:
    """Teardown any state that was previously setup with a setup_module method."""
    Path("my_spectrum_capture").rmdir()


def test_basic_usage() -> None:
    """Runs a basic usage run. Update this code to reflect readme changes."""

    # NOTE: the SDS client-server interaction is stateless, so it is
    #   not recommended to have multiple clients writing to the same
    #   locations simultaneously, as they may overrule each other
    #   and cause loss of data.
    sds = Client(
        host="sds.crc.nd.edu",
        # env_file=Path(".env"),  # default
        # env_config={"SDS_SECRET_TOKEN": "my-custom-token"},  # overrides
    )

    # when in dry-run, no changes are made to the SDS or the local filesystem
    sds.dry_run = True

    # authenticate using either the token from
    # the .env file or in the config passed in
    sds.authenticate()

    # upload all files in a directory to the SDS
    reference_name: str = "my_spectrum_capture"
    local_dir: Path = Path(reference_name)
    sds.upload(
        local_dir,  # may be a single file or a directory
        sds_path=reference_name,  # files will be created under this virtual directory
        verbose=True,  # shows a progress bar (default)
    )

    # download all files in a directory from the SDS
    local_downloads: Path = Path("sds-downloads")
    sds.download(
        sds_path=reference_name,  # files will be downloaded from this virtual directory
        to=local_downloads,  # download to this location; will be created if needed
        overwrite=False,  # do not overwrite local existing files (default)
        verbose=True,  # shows a progress bar (default)
    )


def main() -> None:
    enable_logging()
    test_basic_usage()


if __name__ == "__main__":
    main()



"""Tests for the client module."""

import re
from enum import IntEnum
from pathlib import Path
from typing import Any

import pytest
from spectrumx.client import SDSConfig


class LogLevels(IntEnum):
    """Log levels for testing."""

    TRACE = 5
    DEBUG = 10
    INFO = 20
    WARNING = 30
    ERROR = 40
    CRITICAL = 50


def test_load_config_env_file_not_found(caplog: pytest.LogCaptureFixture) -> None:
    """When the env file is not found, a warning must be logged."""

    non_existent_file = Path("non_existent.env")
    caplog.set_level(LogLevels.WARNING)
    SDSConfig(env_file=non_existent_file)

    assert "file missing" in caplog.text.lower()


def test_load_config_file_found(caplog: pytest.LogCaptureFixture) -> None:
    """A valid env file must be found and loaded."""

    filename = "existing.env"
    existing_file = Path(filename)
    existing_file.touch()
    caplog.set_level(LogLevels.DEBUG)
    SDSConfig(env_file=existing_file, verbose=True)

    pattern = f"found.*{filename}".lower()
    assert re.search(pattern, caplog.text.lower())
    existing_file.unlink()


def test_load_config_from_args(sample_config: dict[str, Any]) -> None:
    """Ensures loading config from arguments works."""

    config = SDSConfig(env_file=Path(".env.example"), env_config=sample_config)

    assert config.api_key == sample_config["SDS_SECRET_TOKEN"]
    assert config.timeout == sample_config["HTTP_TIMEOUT"]


def test_show_config_does_not_expose_api_key(
    sample_config: dict[str, Any],
    caplog: pytest.LogCaptureFixture,
    capsys: pytest.CaptureFixture[str],
) -> None:
    """Makes sure the API key is not exposed by show_config()."""

    config = SDSConfig(env_file=Path(".env.example"), env_config=sample_config)
    config.show_config()

    # this all_outputs approach should work for both `make test` and `make test-verbose`
    captured = capsys.readouterr()
    all_outputs = (captured.out + captured.err + caplog.text).lower()

    assert "api_key" in all_outputs
    assert sample_config["SDS_SECRET_TOKEN"] not in all_outputs


def test_warning_for_unrecognized_config(
    sample_config: dict[str, Any], caplog: pytest.LogCaptureFixture
) -> None:
    """A warning must be logged for unrecognized configuration."""

    unknown_attr = "UNRECOGNIZED_CONFIG"
    sample_config[unknown_attr] = "value"
    caplog.set_level(LogLevels.WARNING)
    SDSConfig(env_file=Path(".env.example"), env_config=sample_config)

    assert "not recognized" in caplog.text.lower()
    assert unknown_attr in caplog.text



"""
Tests for the high-level usage of SpectrumX client.
"""

# pylint: disable=redefined-outer-name
import uuid as uuidlib
from datetime import date
from datetime import datetime
from pathlib import Path

import pytest
from responses import RequestsMock
from spectrumx import Client
from spectrumx.errors import AuthError
from spectrumx.gateway import Endpoints


# --------
# FIXTURES
# --------
@pytest.fixture
def client() -> Client:
    """Fixture to create a Client instance for testing."""
    return Client(host="sds.crc.nd.edu")


@pytest.fixture
def responses_dry_run(responses: RequestsMock) -> RequestsMock:
    """Fixture to mock responses for dry run mode."""
    responses.assert_all_requests_are_fired = False
    return responses


# ------------------------
# TESTS FOR AUTHENTICATION
# ------------------------
def test_authentication_200_succeeds(client: Client, responses: RequestsMock) -> None:
    """Given a successful auth response, the client must be authenticated."""
    responses.get(
        client.base_url + "/auth",
        body="{}",
        status=200,
        content_type="application/json",
    )
    # disable the dry run mode for this,
    # since we're testing the actual request
    client.dry_run = False
    client.authenticate()
    assert client.is_authenticated


def test_authentication_401_fails(client: Client, responses: RequestsMock) -> None:
    """Given a failed auth response, the client must raise AuthError."""
    responses.get(
        client.base_url + "/auth",
        body="{}",
        status=401,
        content_type="application/json",
    )
    client.dry_run = False  # to test the actual request
    with pytest.raises(AuthError):
        client.authenticate()


# -------------------------
# TESTS FOR FILE OPERATIONS
# -------------------------
def test_get_file_by_id(client: Client, responses: RequestsMock) -> None:
    """Given a file ID, the client must return the file."""
    uuid = uuidlib.uuid4()
    url: str = client.base_url + Endpoints.FILES + f"/{uuid.hex}"
    responses.get(
        url=url,
        status=200,
        json={
            "created_at": "2021-10-01T12:00:00",
            "directory": "/my/files/are/here/",
            "expiration_date": "2021-10-01",
            "media_type": "text/plain",
            "name": "file.txt",
            "permissions": "rw-rw-r--",
            "size": 321,
            "updated_at": "2021-10-01T12:00:00",
            "uuid": uuid.hex,
        },
    )
    client.dry_run = False  # to test the actual request
    file_sample = client.get_file(file_uuid=uuid.hex)
    assert file_sample.uuid == uuid


def test_file_get_returns_valid(
    client: Client,
) -> None:
    """The get_file method must return a valid File instance.

    Note the file may not exist locally, but the File instance can still be valid.
        That is the case for dry-runs - which use sample files - and for files fetched
        from the server that have not yet been downloaded.
    """

    file_id = uuidlib.uuid4()
    file_size = 888
    client.dry_run = True
    assert client.dry_run is True, "Dry run must be enabled for this test."
    file_sample = client.get_file(file_uuid=file_id.hex)
    assert file_sample.is_sample is True, "The file must be a sample file"  # pyright: ignore[reportPrivateUsage]
    assert file_sample.uuid is not None
    assert file_sample.name.startswith("dry-run-")
    assert file_sample.media_type == "text/plain"
    assert file_sample.size == file_size
    assert file_sample.directory == Path("./sds-files/dry-run/")
    assert file_sample.permissions == "rw-rw-r--"
    assert isinstance(file_sample.created_at, datetime)
    assert isinstance(file_sample.updated_at, datetime)
    assert isinstance(file_sample.expiration_date, date)
    assert file_sample.expiration_date > file_sample.created_at.date()


def test_file_upload(client: Client, temp_file_with_text_contents: Path) -> None:
    """The upload_file method must return a valid File instance."""
    test_file_size = temp_file_with_text_contents.stat().st_size
    client.dry_run = True
    assert client.dry_run is True, "Dry run must be enabled for this test."
    file_sample = client.upload_file(
        file_path=temp_file_with_text_contents,
        sds_path=Path("/my/upload/location"),
    )
    assert (
        file_sample.is_sample is False
    ), "The file must be a real file on disk (not a sample), even for this test"  # pyright: ignore[reportPrivateUsage]
    assert (
        temp_file_with_text_contents == file_sample.local_path
    ), "Local path does not match"
    assert file_sample.uuid is None, "A local file must not have a UUID"
    assert file_sample.name is not None, "Expected a file name"
    assert file_sample.media_type == "text/plain", "Expected media type 'text/plain'"
    assert file_sample.size == test_file_size, "Expected the test file to be 4030 bytes"
    assert "/tmp/pytest-" in str(  # noqa: S108
        file_sample.local_path
    ), "Expected the temp file directory"
    assert file_sample.directory == Path("/my/upload/location")
    assert file_sample.permissions == "rw-r--r--"
    assert isinstance(file_sample.created_at, datetime)
    assert isinstance(file_sample.updated_at, datetime)
    assert (
        file_sample.expiration_date is None
    ), "Local files should not have an expiration date"


# ----------------------
# TESTS FOR DRY-RUN MODE
# ----------------------
class DryModeAssertionError(AssertionError):
    """Raised in test when a request is made in dry run mode."""


def test_dry_run_setter(client: Client, responses_dry_run: RequestsMock) -> None:
    """Makes sure setter works.

    NOTE: Test behavior of this setter might differ from actual \
        one in early releases. See notes in the setter code.
    """
    client.dry_run = False
    assert client.dry_run is False, "Dry-run setter failed."
    client.dry_run = True
    assert client.dry_run is True, "Dry-run setter failed."


def test_dry_run_enabled_by_default(client: Client) -> None:
    """Dry-run mode must be enabled by default."""
    assert client.dry_run is True, "Dry-run must be enabled by default."


def test_dry_auth_does_not_request(
    client: Client, responses_dry_run: RequestsMock
) -> None:
    """When in dry mode, the client must not make any requests."""
    responses_dry_run.add_callback(
        responses_dry_run.GET,
        url=client.base_url + "/auth",
        callback=lambda _: DryModeAssertionError(
            "No requests must be made in dry run mode"
        ),
    )
    client.dry_run = True
    assert client.dry_run is True, "Dry-run setter failed."
    client.authenticate()


def test_dry_file_upload_does_not_request(
    client: Client, responses_dry_run: RequestsMock, temp_file_with_text_contents: Path
) -> None:
    """When in dry run mode, the upload method must not make any requests."""
    responses_dry_run.add_callback(
        responses_dry_run.POST,
        url=client.base_url + Endpoints.FILES,
        callback=lambda _: DryModeAssertionError(
            "No requests must be made in dry run mode"
        ),
    )
    client.dry_run = True
    assert client.dry_run is True, "Dry run must be enabled for this test."
    _file_sample = client.upload_file(
        file_path=temp_file_with_text_contents,
        sds_path=Path("/my/upload/location"),
    )


def test_dry_file_get_does_not_request(
    client: Client, responses_dry_run: RequestsMock
) -> None:
    """When in dry run mode, the get_file method must not make any requests."""

    file_id = uuidlib.uuid4()

    responses_dry_run.add_callback(
        responses_dry_run.GET,
        url=client.base_url + Endpoints.FILES + f"/{file_id.hex}",
        callback=lambda _: DryModeAssertionError(
            "No requests must be made in dry run mode"
        ),
    )

    client.dry_run = True
    assert client.dry_run is True, "Dry run must be enabled for this test."
    _file_sample = client.get_file(file_uuid=file_id.hex)



"""Data models for the SpectrumX Data System SDK."""

from datetime import date
from datetime import datetime
from pathlib import Path

from pydantic import UUID4
from pydantic import BaseModel

from . import utils


class SDSModel(BaseModel):
    """Base class for most models in SDS."""

    uuid: UUID4 | None = None


class File(SDSModel):
    """A file in SDS.

    Attributes:
        # shared attributes with SDS
        uuid:               The unique identifier of the file in SDS.
        created_at:         The timestamp when the file was created.
        directory:          The path to the file, relative to the user storage in SDS.
        expiration_date:    The date when the file will be marked for deletion from SDS.
        media_type:         The MIME type of the file.
        name:               The user-defined name for this file.
        permissions:        The permissions for the file.
        size:               The size of the file in bytes.
        updated_at:         The timestamp when the file was last updated.

        # local attributes:
        is_sample:          Sample files are not written to disk (used in dry-run mode).
        local_path:         The path to the file on the local filesystem.
    """

    created_at: datetime
    directory: Path
    expiration_date: date | None
    media_type: str
    name: str
    permissions: str
    size: int
    updated_at: datetime

    is_sample: bool = False
    local_path: Path | None = None

    @property
    def path(self) -> str:
        """Returns the path to the file, relative to the owner's root on SDS."""
        return f"{self.directory}/{self.name}"

    @property
    def chmod_props(self) -> str:
        """Converts human-readable permissions to chmod properties."""
        utils.validate_file_permission_string(self.permissions)
        one_hot = "".join(["0" if flag == "-" else "1" for flag in self.permissions])
        perm_num = int(one_hot, base=2)
        max_permission_num = 0o777
        assert 0 <= perm_num <= max_permission_num, "Invalid permission number"
        return f"{perm_num:03o}"


class Dataset(SDSModel):
    """A dataset in SDS (collection of files)."""

    # TODO: Implement this model.


class Capture(SDSModel):
    """A capture in SDS (collection of related RF files)."""

    # TODO: Implement this model.


__all__ = [
    "Capture",
    "Dataset",
    "File",
    "SDSModel",
]



"""
Tests for the SpectrumX client.
"""


def test_version_exists() -> None:
    """Tests the version of the SpectrumX SDK exists."""
    from spectrumx import __version__

    assert isinstance(__version__, str), "Version number must be a string"


def test_version_sem_ver() -> None:
    """Tests the version number follows semantic versioning."""
    from spectrumx import __version__

    parts = __version__.split(".")
    three_parts = 3
    assert len(parts) == three_parts, "Version number must have three parts"
    assert all(part.isdigit() for part in parts)


def test_version_pyproject() -> None:
    """Tests the version is the same as in pyproject.toml."""
    from importlib.metadata import version

    from spectrumx import __version__

    assert __version__ == version(
        "spectrumx",
    ), "Version number must match pyproject.toml"



"""Custom exceptions for the SDS SDK."""

from typing import Any


class SDSError(Exception):
    """Base class for all SDS errors."""

    def __init__(self, message) -> None:
        self.message = message

    def __str__(self) -> Any:
        return self.message


class AuthError(SDSError):
    """Issue with user authentication against SDS."""


class NetworkError(SDSError):
    """Issue with the client connection to the SDS service."""


class ServiceError(SDSError):
    """Issue with the SDS service."""


class FileError(SDSError):
    """Issue interacting with a file in SDS."""


class CaptureError(SDSError):
    """Issue interacting with a capture in SDS."""


class DatasetError(SDSError):
    """Issue interacting with a dataset in SDS."""


class ExperimentError(SDSError):
    """Issue interacting with an experiment in SDS."""



"""Tests for the models module."""

# pylint: disable=redefined-outer-name

from datetime import date
from datetime import datetime
from datetime import timedelta
from pathlib import Path
from typing import Any

import pytest
from pytz import timezone
from spectrumx.models import File


@pytest.fixture
def file_properties() -> dict[str, Any]:
    """Fixture to create a dictionary of file properties."""
    tz = timezone("UTC")
    return {
        "name": "test_file",
        "directory": Path("/my/files/are/here"),
        "media_type": "text/plain",
        "permissions": "-w-rw-r--",
        "size": 111_222_333,
        "created_at": datetime.now(tz=tz) - timedelta(days=14),
        "updated_at": datetime.now(tz=tz),
        "expiration_date": date.today() + timedelta(days=366),  # noqa: DTZ011
    }


def test_file_created(file_properties: dict[str, Any]) -> None:
    """Test that a file can be created correctly."""
    new_file = File(
        **file_properties,
    )
    for key, value in file_properties.items():
        assert (
            getattr(new_file, key) == value
        ), f"{key} does not match: {getattr(new_file, key)} != {value}"


def test_file_path(file_properties: dict[str, Any]) -> None:
    """Test that the path property returns the correct path."""
    new_file = File(
        **file_properties,
    )
    assert new_file.path == str(
        Path(file_properties["directory"]) / file_properties["name"],
    )


def test_chmod_props(file_properties: dict[str, Any]) -> None:
    """Test that the chmod_props property returns the correct chmod properties."""
    new_file = File(
        **file_properties,
    )
    assert new_file.chmod_props == "264"  # 264 (octal) <=> '-w-rw-r--'
